//+------------------------------------------------------------------+
//|                                                   Pattern Helper |
//|                                    Copyright (c) 2012, MandySoft |
//+------------------------------------------------------------------+
// http://www.candlesticker.com/Cs59.asp
// http://stockcharts.com/school/doku.php?id=chart_school:chart_analysis:introduction_to_candlesticks
#property copyright "Copyright  2012, MandySoft"
#property link      "http://www.2p.cz/en/forex_metatrader"
// Version: 0.1
/* Revision:
  v0.1 
    - initial version
*/


#property library

#include <PatternDef.mqh>

#include <Debug.mqh>
#include <stdlib.mqh>
                     
#define NUM_CANDLES 3    // able to calculate formation up to NUM_CANDLES
#define MA_BARS 10       // to calculate trend

bool PatternHelperOpen(bool debug) {
  if (debug) {
    DebugOpen("", "PatternHelper.#.dbg");
  }
  return (true);
}

void PatternHelperClose() {
  DebugClose();
  return;
}

string GetPatternName(int pattern) {
  switch (pattern) {
  case PATTERN1_WHITE_CANDLE:        return("White candle"); break;
  case PATTERN1_SHORT_WHITE_CANDLE:  return("Short white candle"); break;
  case PATTERN1_WHITE_SPINNING_TOP:  return("White spinning top"); break;
  case PATTERN1_LONG_WHITE_CANDLE:   return("Long white candle"); break;
  case PATTERN1_WHITE_MARUBOZU:      return("White marubozu"); break;
  case PATTERN1_WHITE_CLOSING_MARUBOZU: return("White closing marubozu"); break;
  case PATTERN1_WHITE_OPENING_MARUBOZU: return("White opening marubozu"); break;

  case PATTERN1_BLACK_CANDLE:        return("Black candle"); break;
  case PATTERN1_SHORT_BLACK_CANDLE:  return("Short black candle"); break;
  case PATTERN1_BLACK_SPINNING_TOP:  return("Black spinning top"); break;
  case PATTERN1_LONG_BLACK_CANDLE:   return("Long black candle"); break;
  case PATTERN1_BLACK_MARUBOZU:      return("Black marubozu"); break;
  case PATTERN1_BLACK_CLOSING_MARUBOZU: return("Black closing marubozu"); break;
  case PATTERN1_BLACK_OPENING_MARUBOZU: return("Black opening marubozu"); break;

  case PATTERN1_GREY:                return("Grey candle"); break;
  case PATTERN1_HIGH_WAVE:           return("High wave"); break;

  case PATTERN1_DOJI:                return("Doji"); break;
  case PATTERN1_FOUR_PRICE_DOJI:     return("Four price doji"); break;
  case PATTERN1_LONG_LEGGED_DOJI:    return("Long legged doji"); break;
  case PATTERN1_UMBRELLA:            return("Umbrella"); break;
  case PATTERN1_INVERTED_UMBRELLA:   return("Inverted umbrella"); break;

  case PATTERN2_BULL_HARAMI:        return("Bullish harami"); break;
  case PATTERN2_BULL_HARAMI_CROSS:  return("Bullish harami cross"); break;
  case PATTERN2_BULL_ENGULFING:     return("Bullish engulfing"); break;
  case PATTERN2_PIERCING:           return("Piercing"); break;
  case PATTERN2_BULL_STAR:          return("Bullish star"); break;
  case PATTERN2_BULL_DOJI:          return("Bullish doji"); break;
  case PATTERN2_BULL_LONG_LEGGED_DOJI: return("Bullish long-legged doji"); break;
  case PATTERN2_HAMMER:              return("Hammer"); break;
  case PATTERN2_INVERTED_HAMMER:     return("Inverted hammer"); break;
  case PATTERN2_BULL_DRAGONFLY_DOJI: return("Bullish dragon fly doji"); break;
  case PATTERN2_BULL_GRAVESTONE_DOJI: return("Bullish gravestone doji"); break;

  case PATTERN3_THREE_WHITE_SOLDIERS:return("Three white soldiers"); break;
  case PATTERN3_THREE_OUTSIDE_UP:   return("Three outside up"); break;
  case PATTERN3_THREE_INSIDE_UP:    return("Three inside up"); break;
  case PATTERN3_MORNING_STAR:       return("Morning star"); break;
  case PATTERN3_MORNING_DOJI:       return("Morning doji"); break;

  case PATTERN2_BEAR_HARAMI:        return("Bearish harami"); break;
  case PATTERN2_BEAR_HARAMI_CROSS:  return("Bearish harami cross"); break;
  case PATTERN2_BEAR_ENGULFING:     return("Bearish engulfing"); break;
  case PATTERN2_DARK_CLOUD_COVER:   return("Dark cloud cover"); break;
  case PATTERN2_BEAR_STAR:          return("Bearish star"); break;
  case PATTERN2_BEAR_DOJI:          return("Bearish doji"); break;
  case PATTERN2_BEAR_LONG_LEGGED_DOJI: return("Bearish long-legged doji"); break;
  case PATTERN2_HANGING_MAN:        return("Hanging man"); break;
  case PATTERN2_SHOOTING_STAR:      return("Shooting star"); break;
  case PATTERN2_BEAR_DRAGONFLY_DOJI:  return("Bearish dragon fly doji"); break;
  case PATTERN2_BEAR_GRAVESTONE_DOJI: return("Bearish gravestone doji"); break;

  case PATTERN3_THREE_BLACK_CROWS:  return("Three black crows"); break;
  case PATTERN3_THREE_OUTSIDE_DOWN: return("Three outside down"); break;
  case PATTERN3_THREE_INSIDE_DOWN:  return("Three inside down"); break;
  case PATTERN3_EVENING_STAR:       return("Evening star"); break;
  case PATTERN3_EVENING_DOJI:       return("Evening doji"); break;
  }
  return (0);
}

int RecognizePatterns2(double iO[NUM_CANDLES], double iC[NUM_CANDLES], double iL[NUM_CANDLES], double iH[NUM_CANDLES], double trend[NUM_CANDLES], double longLevel, double shortLevel, double zeroLevel, double trendLevel) {
  int i;
  int pat[NUM_CANDLES];
  int patG[NUM_CANDLES];
  
  // calculate single candles
  for (i = 0; i < NUM_CANDLES; i++) {
    pat[i] = 0;
    patG[i] = 0;       
    
    if (MathAbs(iC[i]-iO[i]) > zeroLevel && MathAbs(iC[i]-iO[i]) < shortLevel &&
      (iH[i]-MathMax(iC[i], iO[i]) >= longLevel || MathMin(iC[i],iO[i])-iL[i] >= longLevel) ) {
      // small body long tails
      patG[i] = PATTERN1_GREY;
      pat[i] = PATTERN1_HIGH_WAVE;  
    } else if (iC[i]-iO[i] > zeroLevel) {
      // white candles
      patG[i] = PATTERN1_WHITE_CANDLE;
      if (iC[i]-iO[i] >= longLevel) {
        // long body
        if (iO[i]-iL[i] <= zeroLevel && iH[i]-iC[i] <= zeroLevel) {
          // no tails
          pat[i] = PATTERN1_WHITE_MARUBOZU;
        } else if (iO[i]-iL[i] <= zeroLevel) {
          // no lower tail
          pat[i] = PATTERN1_WHITE_OPENING_MARUBOZU;
        } else if (iH[i]-iC[i] <= zeroLevel) {
          // no upper tail
          pat[i] = PATTERN1_WHITE_CLOSING_MARUBOZU;
        } else {
          pat[i] = PATTERN1_LONG_WHITE_CANDLE;
        }        
      } else if (iC[i]-iO[i] < shortLevel) {
        if (iH[i]-iC[i] <= iC[i]-iO[i] && iO[i]-iL[i] <= iC[i]-iO[i]) {
          pat[i] = PATTERN1_SHORT_WHITE_CANDLE;
        } else if (iH[i]-iC[i] > iC[i]-iO[i] && iO[i]-iL[i] > iC[i]-iO[i]) {
          pat[i] = PATTERN1_WHITE_SPINNING_TOP;
        }
      } else {
        // normal length
        if (iH[i]-iC[i] <= iC[i]-iO[i] && iO[i]-iL[i] <= iC[i]-iO[i]) {
          // normal shadows
          pat[i] = PATTERN1_WHITE_CANDLE;
        }      
      }
    } else if (iO[i]-iC[i] > zeroLevel) {
      // black candles
      patG[i] = PATTERN1_BLACK_CANDLE;
      if (iO[i]-iC[i] >= longLevel) {
        // long body
        if (iH[i]-iO[i] <= zeroLevel && iC[i]-iL[i] <= zeroLevel) {
          // no tails
          pat[i] = PATTERN1_BLACK_MARUBOZU;
        } else if (iH[i]-iO[i] <= zeroLevel) {
          // no upper tail
          pat[i] = PATTERN1_BLACK_OPENING_MARUBOZU;
        } else if (iC[i]-iL[i] <= zeroLevel) {
          // no lower tail
          pat[i] = PATTERN1_BLACK_CLOSING_MARUBOZU;
        } else {
          pat[i] = PATTERN1_LONG_BLACK_CANDLE;
        }
      } else if (iO[i]-iC[i] < shortLevel) {
        if (iC[i]-iL[i] <= iO[i]-iC[i] && iH[i]-iO[i] <= iO[i]-iC[i]) {
          pat[i] = PATTERN1_SHORT_BLACK_CANDLE;
        } else if (iC[i]-iL[i] > iO[i]-iC[i] && iH[i]-iO[i] > iO[i]-iC[i]) {
          pat[i] = PATTERN1_BLACK_SPINNING_TOP;
        }
      } else {
        // normal length
        if (iC[i]-iL[i] <= iO[i]-iC[i] && iH[i]-iO[i] <= iO[i]-iC[i]) {
          // normal shadows
          pat[i] = PATTERN1_BLACK_CANDLE;
        }      
      }        
    } else {
      // no body candles
      patG[i] = PATTERN1_DOJI;
      pat[i] = PATTERN1_DOJI;
      if (iH[i]-MathMax(iC[i],iO[i]) >= longLevel) {
        // long upper
        if (MathMin(iC[i],iO[i])-iL[i] >= longLevel) {
          // long lower
          pat[i] = PATTERN1_LONG_LEGGED_DOJI;
        } else if (MathMin(iC[i], iO[i])-iL[i] <= zeroLevel) {
          pat[i] = PATTERN1_INVERTED_UMBRELLA;
        }
      } else if (MathMin(iC[i],iO[i])-iL[i] >= longLevel) {
        // long towed tail
        if (iH[i]-MathMax(iC[i],iO[i]) <= zeroLevel) {
          // no upper tail
          pat[i] = PATTERN1_UMBRELLA;        
        }        
      } else if (iH[i]-iL[i] <= 2*zeroLevel) {
        pat[i] = PATTERN1_FOUR_PRICE_DOJI;
      }      
    }        
  }

  
  // double candle patterns
  for (i = 0; i < NUM_CANDLES-1; i++) {
    // single + trend
    int hammer = 0;
    if (patG[i] == PATTERN1_DOJI) {
      if (pat[i] == PATTERN1_UMBRELLA) {
        hammer = 2;
      } else if (pat[i] == PATTERN1_INVERTED_UMBRELLA) {
        hammer = -2;
      } else if (pat[i] == PATTERN1_DOJI) {
        hammer = 4;
      } else if (pat[i] == PATTERN1_LONG_LEGGED_DOJI) {
        hammer = 5;
      }
    } else if (MathAbs(iC[i]-iO[i]) < shortLevel) {
      if (iH[i]-MathMax(iC[i],iO[i]) < zeroLevel &&
        MathAbs(iC[i]-iO[i]) < 2*(MathMin(iC[i],iO[i])-iL[i])) {   //lower tail
        hammer = 1;
      } else if (MathMin(iC[i],iO[i])-iL[i] < zeroLevel &&
        MathAbs(iC[i]-iO[i]) < 2*(iH[i]-MathMax(iC[i],iO[i]))) {   //upper tail
        hammer = -1;
      } else if (pat[i] == PATTERN1_SHORT_WHITE_CANDLE || pat[i] == PATTERN1_SHORT_BLACK_CANDLE) {
        hammer = 3;
      }
    } 
    if (hammer != 0) {  
      if ((trend[i] == EMPTY_VALUE || trend[i] < -trendLevel) &&  // down trade
        MathAbs(iC[i+1]-iO[i+1]) >= shortLevel &&  // TMa: optimization
        patG[i+1] == PATTERN1_BLACK_CANDLE &&
        MathMax(iC[i],iO[i]) < iC[i+1]) {  // bellow trend
        
        switch (hammer) {
        case -2: pat[i] |= PATTERN2_BULL_GRAVESTONE_DOJI; break;
        case -1: pat[i] |= PATTERN2_INVERTED_HAMMER; break;
        case  1: pat[i] |= PATTERN2_HAMMER; break;
        case  2: pat[i] |= PATTERN2_BULL_DRAGONFLY_DOJI; break;
        case  3: pat[i] |= PATTERN2_BULL_STAR; break;
        case  4: pat[i] |= PATTERN2_BULL_DOJI; break;
        case  5: pat[i] |= PATTERN2_BULL_LONG_LEGGED_DOJI; break;
        }
      } else if ((trend[i] == EMPTY_VALUE || trend[i] > trendLevel) &&
        MathAbs(iC[i+1]-iO[i+1]) >= shortLevel &&  // TMa: optimization
        patG[i+1] == PATTERN1_WHITE_CANDLE &&
        MathMin(iC[i],iO[i]) > iC[i+1]) {  // above trend
        switch (hammer) {
        case -2: pat[i] |= PATTERN2_BEAR_GRAVESTONE_DOJI; break;
        case -1: pat[i] |= PATTERN2_SHOOTING_STAR; break;
        case  1: pat[i] |= PATTERN2_HANGING_MAN; break;
        case  2: pat[i] |= PATTERN2_BEAR_DRAGONFLY_DOJI; break;
        case  3: pat[i] |= PATTERN2_BEAR_STAR; break;
        case  4: pat[i] |= PATTERN2_BEAR_DOJI; break;
        case  5: pat[i] |= PATTERN2_BEAR_LONG_LEGGED_DOJI; break;
        }            
      }
      if (pat[i] & PATTERN2_MASK != 0) {
        continue;
      }
    }
    if (patG[i+1] == PATTERN1_WHITE_CANDLE && (trend[i] == EMPTY_VALUE || trend[i+1] > trendLevel)) {
      // any white
      if (patG[i] == PATTERN1_BLACK_CANDLE) {
        // any black
        if (iC[i+1] >= iO[i] && iO[i+1] <= iC[i] &&  // inside earlier
          iC[i+1]-iO[i+1] >= iO[i]-iC[i] &&
          iH[i]-iL[i] < iC[i+1]-iO[i+1]+shortLevel && // The shadows (high/low) of the second candlestick are not necessarily contained within the first body, however it's preferable if they are
          iH[i+1] > iH[i] && iL[i+1] < iL[i] ) {         // smaller
          pat[i] |= PATTERN2_BEAR_HARAMI;
        } else if ((iO[i+1]+iC[i+1])/2 > iC[i] &&
            iO[i] > iC[i+1] /* iH[i+1] */ && iC[i] > iO[i+1] && // Then we see a black body characterized by an open above the high of the previous day on the second day.
            iC[i+1]-iO[i+1] >= longLevel) {
          pat[i] |= PATTERN2_DARK_CLOUD_COVER;
        } else if (iO[i] >= iC[i+1] && iO[i+1]>= iC[i] && iO[i]-iC[i] > iC[i+1]-iO[i+1] && iC[i+1]-iO[i+1] < longLevel &&
          iH[i+1]-iL[i+1] < iO[i]-iC[i]+shortLevel && // The white body does not necessarily engulf the shadows of the black body but totally engulfs the body itself
          iH[i] > iH[i+1] && iL[i] < iL[i+1] ) {
          pat[i] |= PATTERN2_BEAR_ENGULFING;
        }
      } else if (patG[i] == PATTERN1_DOJI) {
        if (iC[i+1] >= MathMax(iC[i], iO[i]) && iO[i+1] <= MathMin(iC[i], iO[i]) &&  // inside earlier (doji is considered as "zero" width)
          iH[i]-iL[i] < iC[i+1]-iO[i+1]+shortLevel && // The shadows (high/low) of the second candlestick are not necessarily contained within the first body, however it's preferable if they are
          iH[i+1] > iH[i] && iL[i+1] < iL[i] ) {         // smaller
          pat[i] |= PATTERN2_BEAR_HARAMI_CROSS;
        }
      }
    
    } else if (patG[i+1] == PATTERN1_BLACK_CANDLE && (trend[i] == EMPTY_VALUE || trend[i+1] < -trendLevel)) {
      // any black
      if (patG[i] == PATTERN1_WHITE_CANDLE) {
      // any white
        if (iO[i+1] >= iC[i] && iC[i+1] <= iO[i] &&  // inside earlier
          iO[i+1]-iC[i+1] >= iC[i]-iO[i] && 
          iH[i]-iL[i] < iO[i+1]-iC[i+1]+shortLevel && // The shadows (high/low) of the second candlestick are not necessarily contained within the first body, however it's preferable if they are
          iH[i+1] > iH[i] && iL[i+1] < iL[i] ) { // smaller 
          pat[i] |= PATTERN2_BULL_HARAMI;
        } else if ((iO[i+1]+iC[i+1])/2 < iC[i] &&
            iO[i] < iC[i+1] /* or iL[i+1] */ && iC[i] < iO[i+1] &&  //  Then we see a long white candlestick whose opening price is below previous days low on the second day
            iO[i+1]-iC[i+1] >= longLevel) {
          pat[i] |= PATTERN2_PIERCING;
        } else if (iC[i] >= iO[i+1] && iC[i+1]>= iO[i] && iC[i]-iO[i] > iO[i+1]-iC[i+1] && iO[i+1]-iC[i+1] < longLevel &&
          iH[i+1]-iL[i+1] < iC[i]-iO[i]+shortLevel && // The white body does not necessarily engulf the shadows of the black body but totally engulfs the body itself
          iH[i] > iH[i+1] && iL[i] < iL[i+1] ) {
          pat[i] |= PATTERN2_BULL_ENGULFING;
        }
        
      } else if (patG[i] == PATTERN1_DOJI) {
        if (iC[i+1] <= MathMax(iC[i], iO[i]) && iO[i+1] >= MathMin(iC[i], iO[i]) &&  // inside earlier (doji is considered as "zero" width)
          iH[i]-iL[i] < iO[i+1]-iC[i+1]+shortLevel && // The shadows (high/low) of the second candlestick are not necessarily contained within the first body, however it's preferable if they are
          iH[i+1] > iH[i] && iL[i+1] < iL[i] ) { // smaller 
          pat[i] |= PATTERN2_BULL_HARAMI_CROSS;
        }
      }
    }           
  } 
    
  // triple candle patterns
  for (i = 0; i < NUM_CANDLES-2; i++) {
    if ((trend[i] == EMPTY_VALUE || trend[i+2] < -trendLevel) &&
        patG[i] == PATTERN1_WHITE_CANDLE &&  // or iC-iO > 1% ??
        patG[i+1] == PATTERN1_WHITE_CANDLE && 
        patG[i+2] == PATTERN1_WHITE_CANDLE &&

        iC[i]-iO[i] >= longLevel &&        
        iC[i+1]-iO[i+1] >= longLevel &&        
        iC[i+2]-iO[i+2] >= longLevel &&        
        
        iC[i] > iC[i+1] && iC[i+1] > iC[i+2] &&
        iO[i] < iC[i+1] && iO[i] > iO[i+1] &&
        iO[i+1] < iC[i+2] && iO[i+1] > iO[i+2] &&
           
        iH[i]-iC[i] < shortLevel &&  // Each consecutive day closes near or at its highs
        iH[i+1]-iC[i+1] < shortLevel &&
        iH[i+2]-iC[i+2] < shortLevel ) {
      pat[i] |= PATTERN3_THREE_WHITE_SOLDIERS;
    } else if ((trend[i] == EMPTY_VALUE || trend[i+2] > trendLevel) &&
        patG[i] == PATTERN1_BLACK_CANDLE &&  // or iC-iO > 1% ??
        patG[i+1] == PATTERN1_BLACK_CANDLE && 
        patG[i+2] == PATTERN1_BLACK_CANDLE &&

        iO[i]-iC[i] >= longLevel &&        
        iO[i+1]-iC[i+1] >= longLevel &&        
        iO[i+2]-iC[i+2] >= longLevel &&        
        
        iC[i] < iC[i+1] && iC[i+1] < iC[i+2] &&
        iO[i] > iC[i+1] && iO[i] < iO[i+1] &&
        iO[i+1] > iC[i+2] && iO[i+1] < iO[i+2] &&
           
        iC[i]-iL[i] < shortLevel &&   // Each consecutive day closes near or at its lows
        iC[i+1]-iL[i+1] < shortLevel &&
        iC[i+2]-iL[i+2] < shortLevel ) {
      pat[i] |= PATTERN3_THREE_BLACK_CROWS;
    } else if (pat[i+1] & PATTERN2_MASK == PATTERN2_BULL_HARAMI &&
        patG[i] == PATTERN1_WHITE_CANDLE && 
        iC[i] > iC[i+1] && iO[i] > iO[i+1]) {
      pat[i] |= PATTERN3_THREE_INSIDE_UP;
    } else if (pat[i+1] & PATTERN2_MASK == PATTERN2_BEAR_HARAMI &&
        patG[i] == PATTERN1_BLACK_CANDLE && 
        iC[i] < iC[i+1] && iO[i] < iO[i+1]) {
      pat[i] |= PATTERN3_THREE_INSIDE_DOWN;
    } else if (pat[i+1] & PATTERN2_MASK == PATTERN2_BULL_ENGULFING &&
        patG[i] == PATTERN1_WHITE_CANDLE && 
        iC[i] > iC[i+1] && iO[i] > iO[i+1]) {
      pat[i] |= PATTERN3_THREE_OUTSIDE_UP;
    } else if (pat[i+1] & PATTERN2_MASK == PATTERN2_BEAR_ENGULFING &&
        patG[i] == PATTERN1_BLACK_CANDLE && 
        iC[i] < iC[i+1] && iO[i] < iO[i+1]) {
      pat[i] |= PATTERN3_THREE_OUTSIDE_DOWN;      
    } else if (pat[i+1] & PATTERN2_MASK == PATTERN2_BULL_STAR &&
        patG[i] == PATTERN1_WHITE_CANDLE &&
        iO[i] > MathMax(iO[i+1], iC[i+1])) {
      pat[i] |= PATTERN3_MORNING_STAR;
    } else if (pat[i+1] & PATTERN2_MASK == PATTERN2_BEAR_STAR &&
        patG[i] == PATTERN1_BLACK_CANDLE &&
        iO[i] < MathMin(iO[i+1], iC[i+1])) {
      pat[i] |= PATTERN3_EVENING_STAR;
    } else if (pat[i+1] & PATTERN2_MASK == PATTERN2_BULL_DOJI &&
        patG[i] == PATTERN1_WHITE_CANDLE &&
        iO[i] > MathMax(iO[i+1], iC[i+1])) {
      pat[i] |= PATTERN3_MORNING_DOJI;
    } else if (pat[i+1] & PATTERN2_MASK == PATTERN2_BEAR_DOJI &&
        patG[i] == PATTERN1_BLACK_CANDLE &&
        iO[i] < MathMin(iO[i+1], iC[i+1])) {
      pat[i] |= PATTERN3_EVENING_DOJI;
    }    
  }   
  Debug("RecognizePatterns", 
    iO[0], iC[0], iL[0], iH[0],
    longLevel, shortLevel, zeroLevel, trendLevel, 
    pat[0] & PATTERN1_MASK, GetPatternName(pat[0] & PATTERN1_MASK), 
    IntegerToHexString(pat[0] & PATTERN2_MASK), GetPatternName(pat[0] & PATTERN2_MASK), 
    IntegerToHexString(pat[0] & PATTERN3_MASK), GetPatternName(pat[0] & PATTERN3_MASK));

  return (pat[0]);
}


int RecognizePatterns(/*string symbol, int timeframe, */ int shift, bool considerTrend) {
  double iO[NUM_CANDLES];
  double iC[NUM_CANDLES];
  double iH[NUM_CANDLES];
  double iL[NUM_CANDLES];
  double trend[NUM_CANDLES];
  double longLevel;
  double shortLevel;
  double zeroLevel;
  double trendLevel;
  int i;
  
  if (shift + NUM_CANDLES + MA_BARS + 1 >= Bars) return (0);  // not enough candles
      
  for (i = 0; i < NUM_CANDLES; i++) {
    iO[i] = Open[shift+i];
    iC[i] = Close[shift+i];
    iH[i] = High[shift+i];
    iL[i] = Low[shift+i];
  }

  double body = 0;
  for (i = 0; i < NUM_CANDLES; i++) {
    trend[i] = 0.0;
    body = 0.0;
    for (int j = 0; j < MA_BARS; j++) {
      trend[i] += Close[j+1+i+shift];
      body += MathAbs(Open[j+1+i+shift]-Close[j+1+i+shift]);      
    }
    trend[i] /= MA_BARS;
    if (considerTrend) {
      trend[i] = iO[i]-trend[i];
    } else {
      trend[i] = EMPTY_VALUE;
    }
    body /= MA_BARS;
  }
  longLevel = body * 1.3;    // long bodies, long tails
  shortLevel = body * 0.5;   // short ...
  zeroLevel = body * 0.03;   // no tails, doji
  trendLevel = shortLevel;

  return (RecognizePatterns2(iO, iC, iL, iH, trend, longLevel, shortLevel, zeroLevel, trendLevel));
  
}

int GetPatternInfo(int pattern) {
  switch (pattern) {
  case PATTERN1_WHITE_CANDLE:           return(1|PATTERN_RELIABILITY_LOW); break;
  case PATTERN1_SHORT_WHITE_CANDLE:     return(1|PATTERN_RELIABILITY_LOW); break;
  case PATTERN1_WHITE_SPINNING_TOP:     return(1|PATTERN_RELIABILITY_LOW); break;
  case PATTERN1_LONG_WHITE_CANDLE:      return(1|PATTERN_BULLISH|PATTERN_REVERSAL|PATTERN_CONTINUATION|PATTERN_RELIABILITY_LOW); break;
  case PATTERN1_WHITE_MARUBOZU:         return(1|PATTERN_BULLISH|PATTERN_REVERSAL|PATTERN_CONTINUATION|PATTERN_RELIABILITY_LOW); break;
  case PATTERN1_WHITE_CLOSING_MARUBOZU: return(1|PATTERN_BULLISH|PATTERN_REVERSAL|PATTERN_CONTINUATION|PATTERN_RELIABILITY_LOW); break;
  case PATTERN1_WHITE_OPENING_MARUBOZU: return(1|PATTERN_BULLISH|PATTERN_REVERSAL|PATTERN_CONTINUATION|PATTERN_RELIABILITY_LOW); break;

  case PATTERN1_BLACK_CANDLE:           return(1|PATTERN_RELIABILITY_LOW); break;
  case PATTERN1_SHORT_BLACK_CANDLE:     return(1|PATTERN_RELIABILITY_LOW); break;
  case PATTERN1_BLACK_SPINNING_TOP:     return(1|PATTERN_RELIABILITY_LOW); break;
  case PATTERN1_LONG_BLACK_CANDLE:      return(1|PATTERN_BEARISH|PATTERN_REVERSAL|PATTERN_CONTINUATION|PATTERN_RELIABILITY_LOW); break;
  case PATTERN1_BLACK_MARUBOZU:         return(1|PATTERN_BEARISH|PATTERN_REVERSAL|PATTERN_CONTINUATION|PATTERN_RELIABILITY_LOW); break;
  case PATTERN1_BLACK_CLOSING_MARUBOZU: return(1|PATTERN_BEARISH|PATTERN_REVERSAL|PATTERN_CONTINUATION|PATTERN_RELIABILITY_LOW); break;
  case PATTERN1_BLACK_OPENING_MARUBOZU: return(1|PATTERN_BEARISH|PATTERN_REVERSAL|PATTERN_CONTINUATION|PATTERN_RELIABILITY_LOW); break;

  case PATTERN1_GREY:                   return(1|PATTERN_RELIABILITY_LOW); break;
  case PATTERN1_HIGH_WAVE:              return(1|PATTERN_RELIABILITY_LOW); break;

  case PATTERN1_DOJI:                   return(1|PATTERN_RELIABILITY_LOW); break;
  case PATTERN1_FOUR_PRICE_DOJI:        return(1|PATTERN_RELIABILITY_LOW); break;
  case PATTERN1_LONG_LEGGED_DOJI:       return(1|PATTERN_RELIABILITY_MEDIUM); break;
  case PATTERN1_UMBRELLA:               return(1|PATTERN_RELIABILITY_MEDIUM); break;
  case PATTERN1_INVERTED_UMBRELLA:      return(1|PATTERN_RELIABILITY_MEDIUM); break;

  case PATTERN2_BULL_HARAMI:            return(2|PATTERN_BULLISH|PATTERN_REVERSAL|PATTERN_TREND_DOWN|PATTERN_RELIABILITY_LOW); break;
  case PATTERN2_BULL_HARAMI_CROSS:      return(2|PATTERN_BULLISH|PATTERN_REVERSAL|PATTERN_TREND_DOWN|PATTERN_RELIABILITY_MEDIUM); break;
  case PATTERN2_BULL_ENGULFING:         return(2|PATTERN_BULLISH|PATTERN_REVERSAL|PATTERN_TREND_DOWN|PATTERN_RELIABILITY_MEDIUM); break;
  case PATTERN2_PIERCING:               return(2|PATTERN_BULLISH|PATTERN_REVERSAL|PATTERN_TREND_DOWN|PATTERN_RELIABILITY_HIGH); break;
  case PATTERN2_BULL_STAR:              return(2|PATTERN_BULLISH|PATTERN_REVERSAL|PATTERN_TREND_DOWN|PATTERN_RELIABILITY_MEDIUM); break;
  case PATTERN2_BULL_DOJI:              return(2|PATTERN_BULLISH|PATTERN_REVERSAL|PATTERN_TREND_DOWN|PATTERN_RELIABILITY_MEDIUM); break;
  case PATTERN2_BULL_LONG_LEGGED_DOJI:  return(2|PATTERN_BULLISH|PATTERN_REVERSAL|PATTERN_TREND_DOWN|PATTERN_RELIABILITY_MEDIUM); break;
  case PATTERN2_HAMMER:                 return(2|PATTERN_BULLISH|PATTERN_REVERSAL|PATTERN_TREND_DOWN|PATTERN_RELIABILITY_LOW); break;
  case PATTERN2_INVERTED_HAMMER:        return(2|PATTERN_BULLISH|PATTERN_REVERSAL|PATTERN_TREND_DOWN|PATTERN_RELIABILITY_LOW); break;
  case PATTERN2_BULL_DRAGONFLY_DOJI:    return(2|PATTERN_BULLISH|PATTERN_REVERSAL|PATTERN_TREND_DOWN|PATTERN_RELIABILITY_MEDIUM); break;
  case PATTERN2_BULL_GRAVESTONE_DOJI:   return(2|PATTERN_BULLISH|PATTERN_REVERSAL|PATTERN_TREND_DOWN|PATTERN_RELIABILITY_MEDIUM); break;

  case PATTERN3_THREE_WHITE_SOLDIERS:   return(3|PATTERN_BULLISH|PATTERN_REVERSAL|PATTERN_TREND_DOWN|PATTERN_RELIABILITY_HIGH); break;
  case PATTERN3_THREE_OUTSIDE_UP:       return(3|PATTERN_BULLISH|PATTERN_REVERSAL|PATTERN_TREND_DOWN|PATTERN_RELIABILITY_HIGH); break;
  case PATTERN3_THREE_INSIDE_UP:        return(3|PATTERN_BULLISH|PATTERN_REVERSAL|PATTERN_TREND_DOWN|PATTERN_RELIABILITY_HIGH); break;
  case PATTERN3_MORNING_STAR:           return(3|PATTERN_BULLISH|PATTERN_REVERSAL|PATTERN_TREND_DOWN|PATTERN_RELIABILITY_HIGH); break;
  case PATTERN3_MORNING_DOJI:           return(3|PATTERN_BULLISH|PATTERN_REVERSAL|PATTERN_TREND_DOWN|PATTERN_RELIABILITY_HIGH); break;

  case PATTERN2_BEAR_HARAMI:            return(2|PATTERN_BEARISH|PATTERN_REVERSAL|PATTERN_TREND_UP|PATTERN_RELIABILITY_LOW); break;
  case PATTERN2_BEAR_HARAMI_CROSS:      return(2|PATTERN_BEARISH|PATTERN_REVERSAL|PATTERN_TREND_UP|PATTERN_RELIABILITY_MEDIUM); break;
  case PATTERN2_BEAR_ENGULFING:         return(2|PATTERN_BEARISH|PATTERN_REVERSAL|PATTERN_TREND_UP|PATTERN_RELIABILITY_MEDIUM); break;
  case PATTERN2_DARK_CLOUD_COVER:       return(2|PATTERN_BEARISH|PATTERN_REVERSAL|PATTERN_TREND_UP|PATTERN_RELIABILITY_HIGH); break;
  case PATTERN2_BEAR_STAR:              return(2|PATTERN_BEARISH|PATTERN_REVERSAL|PATTERN_TREND_UP|PATTERN_RELIABILITY_MEDIUM); break;
  case PATTERN2_BEAR_DOJI:              return(2|PATTERN_BEARISH|PATTERN_REVERSAL|PATTERN_TREND_UP|PATTERN_RELIABILITY_MEDIUM); break;
  case PATTERN2_BEAR_LONG_LEGGED_DOJI:  return(2|PATTERN_BEARISH|PATTERN_REVERSAL|PATTERN_TREND_UP|PATTERN_RELIABILITY_MEDIUM); break;
  case PATTERN2_HANGING_MAN:            return(2|PATTERN_BEARISH|PATTERN_REVERSAL|PATTERN_TREND_UP|PATTERN_RELIABILITY_LOW); break;
  case PATTERN2_SHOOTING_STAR:          return(2|PATTERN_BEARISH|PATTERN_REVERSAL|PATTERN_TREND_UP|PATTERN_RELIABILITY_LOW); break;
  case PATTERN2_BEAR_DRAGONFLY_DOJI:    return(2|PATTERN_BEARISH|PATTERN_REVERSAL|PATTERN_TREND_UP|PATTERN_RELIABILITY_MEDIUM); break;
  case PATTERN2_BEAR_GRAVESTONE_DOJI:   return(2|PATTERN_BEARISH|PATTERN_REVERSAL|PATTERN_TREND_UP|PATTERN_RELIABILITY_MEDIUM); break;

  case PATTERN3_THREE_BLACK_CROWS:      return(3|PATTERN_BEARISH|PATTERN_REVERSAL|PATTERN_TREND_UP|PATTERN_RELIABILITY_HIGH); break;
  case PATTERN3_THREE_OUTSIDE_DOWN:     return(3|PATTERN_BEARISH|PATTERN_REVERSAL|PATTERN_TREND_UP|PATTERN_RELIABILITY_HIGH); break;
  case PATTERN3_THREE_INSIDE_DOWN:      return(3|PATTERN_BEARISH|PATTERN_REVERSAL|PATTERN_TREND_UP|PATTERN_RELIABILITY_HIGH); break;
  case PATTERN3_EVENING_STAR:           return(3|PATTERN_BEARISH|PATTERN_REVERSAL|PATTERN_TREND_UP|PATTERN_RELIABILITY_HIGH); break;
  case PATTERN3_EVENING_DOJI:           return(3|PATTERN_BEARISH|PATTERN_REVERSAL|PATTERN_TREND_UP|PATTERN_RELIABILITY_HIGH); break;
  }
  
  return (0);
}

bool testResult(string phase, int testN, int pat1, int pat2, int patExpected[5]) {
  bool ret = true;
    if (pat1 & PATTERN1_MASK != pat2 & PATTERN1_MASK) {
      Debug(testN, phase, "Pattern1!= Pattern2", pat1 & PATTERN1_MASK, pat2 & PATTERN1_MASK);
      ret = false;
    }
    if (pat1 & PATTERN1_MASK != patExpected[0]) {
      Debug(testN, phase, "Pattern1", GetPatternName(pat1 & PATTERN1_MASK), GetPatternName(patExpected[0]));
      ret = false;
    }
    if (pat1 & PATTERN2_MASK != patExpected[1]) {
      Debug(testN, phase, "Pattern2-UP", GetPatternName(pat1 & PATTERN2_MASK), GetPatternName(patExpected[1]));
      ret = false;
    }
    if (pat1 & PATTERN3_MASK != patExpected[2]) {
      Debug(testN, phase, "Pattern3-UP", GetPatternName(pat1 & PATTERN3_MASK), GetPatternName(patExpected[2]));
      ret = false;
    }
    if (pat2 & PATTERN2_MASK != patExpected[3]) {
      Debug(testN, phase, "Pattern2-DOWN", GetPatternName(pat2 & PATTERN2_MASK), GetPatternName(patExpected[3]));
      ret = false;
    }
    if (pat2 & PATTERN3_MASK != patExpected[4]) {
      Debug(testN, phase, "Pattern3-DOWN", GetPatternName(pat2 & PATTERN3_MASK), GetPatternName(patExpected[4]));
      ret = false;
    }        

  return(ret);
}

// unit test of paterns
int PatternHelperTest() {
#define NUM_FLDS 9
// only white and grey patterns. Symmetric (black) patterns are automatically inverted and tested too.
  int testData[] = { // O, C, L, H, PATTERN1, trend up PATTERN2, trend up PATTERN3, trend down PATTERN2, trend down PATTERN3 */
    1, 2, 1, 2, 0, 0, 0, 0, 0,
    1, 2, 1, 2, 0, 0, 0, 0, 0,
    1, 100, 1, 100, PATTERN1_WHITE_MARUBOZU, 0, 0, 0, 0,
    1, 90, 1, 100, PATTERN1_WHITE_OPENING_MARUBOZU, 0, 0, 0, 0,
    10, 100, 1, 100, PATTERN1_WHITE_CLOSING_MARUBOZU, 0, 0, 0, 0,
    10, 90, 1, 100, PATTERN1_LONG_WHITE_CANDLE, 0, 0, 0, 0,
    5, 15, 1, 20, PATTERN1_SHORT_WHITE_CANDLE, 0, 0, 0, 0,
    50, 55, 40, 65, PATTERN1_WHITE_SPINNING_TOP, 0, 0, 0, 0,
    20, 50, 1, 60, PATTERN1_WHITE_CANDLE, 0, 0, 0, 0,    
        
    100, 105, 1, 200, PATTERN1_HIGH_WAVE, 0, 0, 0, 0,
    100, 105, 1, 200, PATTERN1_HIGH_WAVE, 0, 0, 0, 0,
    100, 105, 1, 120, PATTERN1_HIGH_WAVE, 0, 0, 0, 0,
    100, 105, 80, 200, PATTERN1_HIGH_WAVE, 0, 0, 0, 0,

    49, 50, 40, 60, PATTERN1_DOJI, 0, 0, 0, 0,
    51, 50, 49, 51, PATTERN1_FOUR_PRICE_DOJI, 0, 0, 0, 0,
    100, 102, 1, 200, PATTERN1_LONG_LEGGED_DOJI, 0, 0, 0, 0,
    71, 70, 1, 72, PATTERN1_UMBRELLA, 0, 0, 0, 0,
    1, 2, 1, 72, PATTERN1_INVERTED_UMBRELLA, 0, 0, 0, 0,
    
    70, 20, 10, 80, PATTERN1_BLACK_CANDLE, 0, 0, 0, 0,
    30, 60, 25, 68, PATTERN1_WHITE_CANDLE, 0, 0, PATTERN2_BULL_HARAMI, 0, 
    70, 20, 10, 80, PATTERN1_BLACK_CANDLE, PATTERN2_BEAR_ENGULFING, 0, 0, 0,
    30, 60, 5, 68, PATTERN1_WHITE_CANDLE, 0, 0, 0, 0, 
    70, 20, 10, 80, PATTERN1_BLACK_CANDLE, 0, 0, 0, 0,
    48, 50, 30, 60, PATTERN1_DOJI, 0, 0, PATTERN2_BULL_HARAMI_CROSS, 0, 
    50, 30, 25, 55, PATTERN1_BLACK_CANDLE, 0, 0, 0, 0,
    25, 60, 20, 65, PATTERN1_WHITE_CANDLE, 0, 0, PATTERN2_BULL_ENGULFING, 0, 
    50, 30, 25, 55, PATTERN1_BLACK_CANDLE, PATTERN2_BEAR_HARAMI, 0, 0, 0,
    26, 60, 20, 65, PATTERN1_WHITE_CANDLE, 0, 0, PATTERN2_BULL_ENGULFING, 0, 
    80, 20, 15, 85, PATTERN1_LONG_BLACK_CANDLE, PATTERN2_BEAR_ENGULFING, 0, 0, 0,
    10, 72, 6, 76, PATTERN1_LONG_WHITE_CANDLE, 0, 0, PATTERN2_PIERCING, 0, 
    80, 20, 15, 85, PATTERN1_LONG_BLACK_CANDLE, PATTERN2_DARK_CLOUD_COVER, 0, 0, 0,
    12, 40, 11, 65, PATTERN1_WHITE_CANDLE, 0, 0, 0, 0, 
    80, 30, 25, 85, PATTERN1_BLACK_CANDLE, 0, 0, 0, 0,
    15, 22, 14, 25, PATTERN1_SHORT_WHITE_CANDLE, 0, 0, PATTERN2_BULL_STAR, 0, 
    80, 30, 25, 85, PATTERN1_BLACK_CANDLE, 0, 0, 0, 0,
    19, 22, 14, 25, PATTERN1_DOJI, 0, 0, PATTERN2_BULL_DOJI, 0, 
    150, 130, 125, 155, PATTERN1_BLACK_CANDLE, 0, 0, 0, 0,
    80, 78, 10, 141, PATTERN1_LONG_LEGGED_DOJI, 0, 0, PATTERN2_BULL_LONG_LEGGED_DOJI, 0, 
    100, 60, 55, 110, PATTERN1_BLACK_CANDLE, 0, 0, 0, 0,
    55, 50, 10, 56, 0, 0, 0, PATTERN2_HAMMER, 0, 
    100, 60, 55, 110, PATTERN1_BLACK_CANDLE, 0, 0, 0, 0,
    30, 55, 10, 56, PATTERN1_WHITE_CANDLE, 0, 0, 0, 0, 
    100, 60, 55, 110, PATTERN1_BLACK_CANDLE, 0, 0, 0, 0,
    55, 50, 49, 70, 0, 0, 0, PATTERN2_INVERTED_HAMMER, 0, 
    100, 80, 70, 110, PATTERN1_BLACK_CANDLE, 0, 0, 0, 0,
    70, 68, 5, 71, PATTERN1_UMBRELLA, 0, 0, PATTERN2_BULL_DRAGONFLY_DOJI, 0, 
    100, 80, 70, 110, PATTERN1_BLACK_CANDLE, 0, 0, 0, 0,
    68, 70, 67, 140, PATTERN1_INVERTED_UMBRELLA, 0, 0, PATTERN2_BULL_GRAVESTONE_DOJI, 0, 


    20, 80, 5, 85, PATTERN1_LONG_WHITE_CANDLE, 0, 0, 0, 0,
    30, 90, 25, 95, PATTERN1_LONG_WHITE_CANDLE, 0, 0, 0, 0,
    40, 100, 35, 105, PATTERN1_LONG_WHITE_CANDLE, 0, 0, 0, PATTERN3_THREE_WHITE_SOLDIERS,
    50, 20, 15, 55, PATTERN1_BLACK_CANDLE, 0, 0, 0, 0,
    16, 52, 10, 60, PATTERN1_WHITE_CANDLE, 0, 0, PATTERN2_BULL_ENGULFING, 0,
    50, 100, 40, 120, PATTERN1_WHITE_CANDLE, 0, 0, 0, PATTERN3_THREE_OUTSIDE_UP,
    76, 16, 10, 80, PATTERN1_LONG_BLACK_CANDLE, 0, 0, 0, 0,
    20, 50, 15, 60, PATTERN1_WHITE_CANDLE, 0, 0, PATTERN2_BULL_HARAMI, 0,
    45, 100, 40, 120, PATTERN1_WHITE_CANDLE, 0, 0, 0, PATTERN3_THREE_INSIDE_UP,

    1, 2, 1, 2, PATTERN1_FOUR_PRICE_DOJI, 0, 0, 0, 0,
    
    100, 40, 35, 110, PATTERN1_LONG_BLACK_CANDLE, 0, 0, 0, 0,
    35, 25, 20, 40, PATTERN1_SHORT_BLACK_CANDLE, 0, 0, PATTERN2_BULL_STAR, 0,
    50, 70, 45, 75, PATTERN1_WHITE_CANDLE, 0, 0, 0, PATTERN3_MORNING_STAR,

    1, 2, 1, 2, PATTERN1_FOUR_PRICE_DOJI, 0, 0, 0, 0,

    100, 40, 35, 110, PATTERN1_LONG_BLACK_CANDLE, 0, 0, 0, 0,
    25, 35, 20, 40, PATTERN1_SHORT_WHITE_CANDLE, 0, 0, PATTERN2_BULL_STAR, 0,
    50, 70, 45, 75, PATTERN1_WHITE_CANDLE, 0, 0, 0, PATTERN3_MORNING_STAR,

    1, 2, 1, 2, PATTERN1_FOUR_PRICE_DOJI, 0, 0, 0, 0,
    
    100, 40, 35, 110, PATTERN1_LONG_BLACK_CANDLE, 0, 0, 0, 0,
    25, 25, 20, 40, PATTERN1_DOJI, 0, 0, PATTERN2_BULL_DOJI, 0,
    50, 70, 45, 75, PATTERN1_WHITE_CANDLE, 0, 0, 0, PATTERN3_MORNING_DOJI,
    
    
    1, 2, 1, 2, PATTERN1_FOUR_PRICE_DOJI, 0, 0, 0, 0
  
  };
  bool ret = true;
  int testN = 0;
  for (int i=0; i<ArraySize(testData)-2*NUM_FLDS; i=i+NUM_FLDS) {
    double iO[NUM_CANDLES];
    double iC[NUM_CANDLES];
    double iO2[NUM_CANDLES];
    double iC2[NUM_CANDLES];
    double iL[NUM_CANDLES];
    double iH[NUM_CANDLES];    
    double trend1[NUM_CANDLES];    
    double trend2[NUM_CANDLES];    
    int patExpected[5];
    int n = i;
    int min = 99999;
    int max = 0;
    for (int j=NUM_CANDLES-1; j>=0; j--) {
      
      iO[j] = testData[n]; n++;
      iC[j] = testData[n]; n++;
      iL[j] = testData[n]; n++;
      iH[j] = testData[n]; n++;
      
      max = MathMax(iH[j], max);
      min = MathMin(iH[j], min);
        
      for (int l=0; l<ArraySize(patExpected); l++) {
        patExpected[l] = testData[n]; n++;
      }
      trend1[j] = 50;
      trend2[j] = -50;      
    }
    
    
    if (iL[0] > iO[0] || iL[0] > iC[0] || iH[0] < iO[0] || iH[0] < iC[0]) {
      Alert("Wrong test data: ", testN, ", O:", iO[0], ", C:", iC[0], ", L:", iL[0], ", H:", iH[0]);
    }
    int pat1 = RecognizePatterns2(iO, iC, iL, iH, trend1, 60, 20, 3, 20);
    int pat2 = RecognizePatterns2(iO, iC, iL, iH, trend2, 60, 20, 3, 20);

    for (j=0; j<NUM_CANDLES; j++) {
      // inverse open/close, symmetric by center      
      iO[j] = min+max-iO[j];
      iC[j] = min+max-iC[j];
      iL[j] = min+max-iL[j];
      iH[j] = min+max-iH[j];
    }
    
    int pat3 = RecognizePatterns2(iO, iC, iH, iL, trend2, 60, 20, 3, 20);
    int pat4 = RecognizePatterns2(iO, iC, iH, iL, trend1, 60, 20, 3, 20);
     
    ret = ret && testResult("NOR", testN, pat1, pat2, patExpected);
    
    if (patExpected[0] >= PATTERN1_BLACK_CANDLE && patExpected[0] < 2*PATTERN1_BLACK_CANDLE-PATTERN1_WHITE_CANDLE) {
      patExpected[0] -= PATTERN1_BLACK_CANDLE-PATTERN1_WHITE_CANDLE;
    } else if (patExpected[0] >= PATTERN1_WHITE_CANDLE && patExpected[0] < PATTERN1_BLACK_CANDLE) {
      patExpected[0] += PATTERN1_BLACK_CANDLE-PATTERN1_WHITE_CANDLE;
    } 
    switch (patExpected[0]) {
    case PATTERN1_UMBRELLA:
      patExpected[0] = PATTERN1_INVERTED_UMBRELLA; break;
    case PATTERN1_INVERTED_UMBRELLA:
      patExpected[0] = PATTERN1_UMBRELLA; break;
    }
      
    for (int k = 1; k<=3; k=k+2) {
      if (patExpected[k] & 0x0F00 != 0) {
        patExpected[k] = patExpected[k] << 4;
      } else if (patExpected[k] & 0x0F000 != 0) {
        patExpected[k] = patExpected[k] >> 4;
      }
      if (patExpected[k+1] & 0x0F0000 != 0) {
        patExpected[k+1] = patExpected[k+1] << 4;
      } else if (patExpected[k+1] & 0x0F00000 != 0) {
        patExpected[k+1] = patExpected[k+1] >> 4;
      }
    }
    ret = ret && testResult("INV", testN, pat3, pat4, patExpected);
         
    testN++;
  }    
  return (ret);
}


